"""KG Morpho Feature tool."""

import logging
from typing import Any, Literal, Type

from langchain_core.tools import ToolException
from pydantic import BaseModel, Field, model_validator

from neuroagent.tools.base_tool import BaseToolOutput, BasicTool
from neuroagent.utils import get_descendants_id

LABEL = Literal[
    "Neurite Max Radial Distance",
    "Number Of Sections",
    "Number Of Bifurcations",
    "Number Of Leaves",
    "Total Length",
    "Total Area",
    "Total Volume",
    "Section Lengths",
    "Section Term Lengths",
    "Section Bif Lengths",
    "Section Branch Orders",
    "Section Bif Branch Orders",
    "Section Term Branch Orders",
    "Section Path Distances",
    "Section Taper Rates",
    "Local Bifurcation Angles",
    "Remote Bifurcation Angles",
    "Partition Asymmetry",
    "Partition Asymmetry Length",
    "Sibling Ratios",
    "Diameter Power Relations",
    "Section Radial Distances",
    "Section Term Radial Distances",
    "Section Bif Radial Distances",
    "Terminal Path Lengths",
    "Section Volumes",
    "Section Areas",
    "Section Tortuosity",
    "Section Strahler Orders",
    "Soma Surface Area",
    "Soma Radius",
    "Soma Number Of Points",
    "Morphology Max Radial Distance",
    "Number Of Sections Per Neurite",
    "Total Length Per Neurite",
    "Total Area Per Neurite",
    "Total Height",
    "Total Width",
    "Total Depth",
    "Number Of Neurites",
]

STATISTICS = {
    "Morphology Max Radial Distance": ["raw"],
    "Neurite Max Radial Distance": ["raw"],
    "Number Of Sections": ["raw"],
    "Number Of Bifurcations": ["raw"],
    "Number Of Leaves": ["raw"],
    "Total Length": ["raw"],
    "Total Area": ["raw"],
    "Total Volume": ["raw"],
    "Section Lengths": ["min", "max", "median", "mean", "std"],
    "Section Term Lengths": ["min", "max", "median", "mean", "std"],
    "Section Bif Lengths": ["min", "max", "median", "mean", "std"],
    "Section Branch Orders": ["min", "max", "median", "mean", "std"],
    "Section Bif Branch Orders": ["min", "max", "median", "mean", "std"],
    "Section Term Branch Orders": ["min", "max", "median", "mean", "std"],
    "Section Path Distances": ["min", "max", "median", "mean", "std"],
    "Section Taper Rates": ["min", "max", "median", "mean", "std"],
    "Local Bifurcation Angles": ["min", "max", "median", "mean", "std"],
    "Remote Bifurcation Angles": ["min", "max", "median", "mean", "std"],
    "Partition Asymmetry": ["min", "max", "median", "mean", "std"],
    "Partition Asymmetry Length": ["min", "max", "median", "mean", "std"],
    "Sibling Ratios": ["min", "max", "median", "mean", "std"],
    "Diameter Power Relations": ["min", "max", "median", "mean", "std"],
    "Section Radial Distances": ["min", "max", "median", "mean", "std"],
    "Section Term Radial Distances": ["min", "max", "median", "mean", "std"],
    "Section Bif Radial Distances": ["min", "max", "median", "mean", "std"],
    "Terminal Path Lengths": ["min", "max", "median", "mean", "std"],
    "Section Volumes": ["min", "max", "median", "mean", "std"],
    "Section Areas": ["min", "max", "median", "mean", "std"],
    "Section Tortuosity": ["min", "max", "median", "mean", "std"],
    "Section Strahler Orders": ["min", "max", "median", "mean", "std"],
    "Soma Surface Area": ["raw"],
    "Soma Radius": ["raw"],
    "Number Of Sections Per Neurite": ["min", "max", "median", "mean", "std"],
    "Total Length Per Neurite": ["min", "max", "median", "mean", "std"],
    "Total Area Per Neurite": ["min", "max", "median", "mean", "std"],
    "Total Height": ["raw"],
    "Total Width": ["raw"],
    "Total Depth": ["raw"],
    "Number Of Neurites": ["raw"],
    "Soma Number Of Points": ["N"],
}

logger = logging.getLogger(__name__)


class FeatRangeInput(BaseModel):
    """Features Range input class."""

    min_value: float | int | None = None
    max_value: float | int | None = None


class FeatureInput(BaseModel):
    """Class defining the scheme of inputs the agent should use for the features."""

    label: LABEL
    compartment: (
        Literal["Axon", "BasalDendrite", "ApicalDendrite", "NeuronMorphology", "Soma"]
        | None
    ) = Field(
        default=None,
        description=(
            "Compartment of the cell. Leave as None if not explicitely stated by the"
            " user"
        ),
    )
    feat_range: FeatRangeInput | None = None

    @model_validator(mode="before")
    @classmethod
    def check_if_list(cls, data: Any) -> dict[str, str | list[float | int] | None]:
        """Validate that the values passed to the constructor are a dictionary."""
        if isinstance(data, list) and len(data) == 1:
            data_dict = data[0]
        else:
            data_dict = data
        return data_dict


class InputKGMorphoFeatures(BaseModel):
    """Inputs of the knowledge graph API when retrieving features of morphologies."""

    brain_region_id: str = Field(description="ID of the brain region of interest.")
    features: FeatureInput = Field(
        description="""Definition of the feature and values expected by the user.
        The input consists of a dictionary with three keys. The first one is the label (or name) of the feature specified by the user.
        The second one is the compartment in which the feature is calculated. It MUST be None if not explicitly specified by the user.
        The third one consists of a min_value and a max_value which encapsulate the range of values the user expects for this feature. It can also be None if not specified by the user.
        For instance, if the user asks for a morphology with an axon section volume between 1000 and 5000µm, the corresponding tuple should be: {label: 'Section Volumes', compartment: 'Axon', feat_range: FeatRangeInput(min_value=1000, max_value=5000)}.""",
    )


class KGMorphoFeatureOutput(BaseToolOutput):
    """Output schema for the knowledge graph API."""

    brain_region_id: str
    brain_region_label: str | None = None

    morphology_id: str
    morphology_name: str | None = None

    features: dict[str, str]


class KGMorphoFeatureTool(BasicTool):
    """Class defining the Knowledge Graph logic."""

    name: str = "kg-morpho-feature-tool"
    description: str = """Searches a neuroscience based knowledge graph to retrieve neuron morphology features based on a brain region of interest.
    Use this tool if and only if the user specifies explicitely certain features of morphology, and potentially the range of values expected.
    Requires a 'brain_region_id' and a dictionary with keys 'label' (and optionally 'compartment' and 'feat_range') describing the feature(s) specified by the user.
    The morphology ID is in the form of an HTTP(S) link such as 'https://bbp.epfl.ch/neurosciencegraph/data/neuronmorphologies...'.
    The output is a list of morphologies, containing:
    - The brain region ID.
    - The brain region name.
    - The morphology ID.
    - The morphology name.
    - The list of features of the morphology.
    If a given feature has multiple statistics (e.g. mean, min, max, median...), please return only its mean unless specified differently by the user."""
    metadata: dict[str, Any]
    args_schema: Type[BaseModel] = InputKGMorphoFeatures

    def _run(self) -> None:
        """Not defined yet."""
        pass

    async def _arun(
        self,
        brain_region_id: str,
        features: FeatureInput,
    ) -> list[KGMorphoFeatureOutput] | dict[str, str]:
        """Run the tool async.

        Parameters
        ----------
        brain_region_id
            ID of the brain region of interest (of the form http://api.brain-map.org/api/v2/data/Structure/...)
        features
            Pydantic class describing the features one wants to compute

        Returns
        -------
            list of KGMorphoFeatureOutput to describe the morphology and its features, or an error dict.
        """
        try:
            logger.info(
                f"Entering KG morpho feature tool. Inputs: {brain_region_id=},"
                f" {features=}"
            )
            # Get the descendants of the brain region specified as input
            hierarchy_ids = get_descendants_id(
                brain_region_id, json_path=self.metadata["brainregion_path"]
            )
            logger.info(
                f"Found {len(list(hierarchy_ids))} children of the brain ontology."
            )

            # Get the associated ES query
            entire_query = self.create_query(
                brain_regions_ids=hierarchy_ids, features=features
            )

            # Send the ES query to the KG
            response = await self.metadata["httpx_client"].post(
                url=self.metadata["url"],
                headers={"Authorization": f"Bearer {self.metadata['token']}"},
                json=entire_query,
            )

            return self._process_output(response.json())

        except Exception as e:
            raise ToolException(str(e), self.name)

    def create_query(
        self, brain_regions_ids: set[str], features: FeatureInput
    ) -> dict[str, Any]:
        """Create ES query to query the KG with.

        Parameters
        ----------
        brain_regions_ids
            IDs of the brain region of interest (of the form http://api.brain-map.org/api/v2/data/Structure/...)
        features
            Pydantic class describing the features one wants to compute

        Returns
        -------
            Dict containing the ES query to send to the KG.
        """
        # At least one BR should match in the set of descendants
        conditions = [
            {
                "bool": {
                    "should": [
                        {"term": {"brainRegion.@id.keyword": hierarchy_id}}
                        for hierarchy_id in brain_regions_ids
                    ]
                }
            }
        ]

        # Add condition for the name of the requested feature to be present
        sub_conditions: list[dict[str, Any]] = []
        sub_conditions.append(
            {"term": {"featureSeries.label.keyword": str(features.label)}}
        )

        # Optionally add a constraint on the compartment if specified
        if features.compartment:
            sub_conditions.append(
                {
                    "term": {
                        "featureSeries.compartment.keyword": str(features.compartment)
                    }
                }
            )

        # Optionally add a constraint on the feature values if specified
        if features.feat_range:
            # Get the correct statistic for the feature
            stat = (
                "mean"
                if "mean" in STATISTICS[features.label]
                else "raw"
                if "raw" in STATISTICS[features.label]
                else "N"
            )

            # Add constraint on the statistic type
            sub_conditions.append({"term": {"featureSeries.statistic.keyword": stat}})
            feat_range = [
                features.feat_range.min_value,
                features.feat_range.max_value,
            ]
            # Add constraint on min and/or max value of the feature
            sub_condition = {"range": {"featureSeries.value": {}}}  # type: ignore
            if feat_range[0]:
                sub_condition["range"]["featureSeries.value"]["gte"] = feat_range[0]
            if feat_range[1]:
                sub_condition["range"]["featureSeries.value"]["lte"] = feat_range[1]
            if len(sub_condition["range"]["featureSeries.value"]) > 0:
                sub_conditions.append(sub_condition)

        # Nest the entire constrained query in a nested block
        feature_nested_query = {
            "nested": {
                "path": "featureSeries",
                "query": {"bool": {"must": sub_conditions}},
            }
        }
        conditions.append(feature_nested_query)  # type: ignore

        # Unwrap all of the conditions in the global query
        entire_query = {
            "size": self.metadata["search_size"],
            "track_total_hits": True,
            "query": {
                "bool": {
                    "must": [
                        *conditions,
                        {
                            "term": {
                                "@type.keyword": "https://bbp.epfl.ch/ontologies/core/bmo/NeuronMorphologyFeatureAnnotation"
                            }
                        },
                        {"term": {"deprecated": False}},
                    ]
                }
            },
        }

        return entire_query

    @staticmethod
    def _process_output(output: Any) -> list[KGMorphoFeatureOutput]:
        """Process output.

        Parameters
        ----------
        output
            Raw output of the _arun method, which comes from the KG

        Returns
        -------
            list of KGMorphoFeatureOutput to describe the morphology and its features.
        """
        formatted_output = []
        for morpho in output["hits"]["hits"]:
            morpho_source = morpho["_source"]
            feature_output = {
                f"{dic['compartment']} {dic['label']} ({dic['statistic']})": (
                    f"{dic['value']} ({dic['unit']})"
                )
                for dic in morpho_source["featureSeries"]
            }
            formatted_output.append(
                KGMorphoFeatureOutput(
                    brain_region_id=morpho_source["brainRegion"]["@id"],
                    brain_region_label=morpho_source["brainRegion"].get("label"),
                    morphology_id=morpho_source["neuronMorphology"]["@id"],
                    morphology_name=morpho_source["neuronMorphology"].get("name"),
                    features=feature_output,
                )
            )

        return formatted_output



"""Utils related to object resolving in the KG."""

import asyncio
import logging
import re
from typing import Literal

from httpx import AsyncClient

logger = logging.getLogger(__name__)


SPARQL_QUERY = """
   PREFIX bmc: <https://sbo-nexus-delta.shapes-registry.org/ontologies/core/bmc/>
   PREFIX bmo: <https://sbo-nexus-delta.shapes-registry.org/ontologies/core/bmo/>
   PREFIX bmoutils: <https://sbo-nexus-delta.shapes-registry.org/ontologies/core/bmoutils/>
   PREFIX commonshapes: <https://neuroshapes.org/commons/>
   PREFIX datashapes: <https://neuroshapes.org/dash/>
   PREFIX dc: <http://purl.org/dc/elements/1.1/>
   PREFIX dcat: <http://www.w3.org/ns/dcat#>
   PREFIX dcterms: <http://purl.org/dc/terms/>
   PREFIX mba: <http://api.brain-map.org/api/v2/data/Structure/>
   PREFIX nsg: <https://neuroshapes.org/>
   PREFIX nxv: <https://bluebrain.github.io/nexus/vocabulary/>
   PREFIX oa: <http://www.w3.org/ns/oa#>
   PREFIX obo: <http://purl.obolibrary.org/obo/>
   PREFIX owl: <http://www.w3.org/2002/07/owl#>
   PREFIX prov: <http://www.w3.org/ns/prov#>
   PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
   PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
   PREFIX schema: <http://schema.org/>
   PREFIX sh: <http://www.w3.org/ns/shacl#>
   PREFIX shsh: <http://www.w3.org/ns/shacl-shacl#>
   PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
   PREFIX vann: <http://purl.org/vocab/vann/>
   PREFIX void: <http://rdfs.org/ns/void#>
   PREFIX xml: <http://www.w3.org/XML/1998/namespace/>
   PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
   PREFIX : <https://sbo-nexus-delta.shapes-registry.org/ontologies/core/bmo/>

           CONSTRUCT {{
               ?id a ?type ;
               rdfs:label ?label ;
               skos:prefLabel ?prefLabel ;
               skos:altLabel ?altLabel ;
               skos:definition ?definition;
               rdfs:subClassOf ?subClassOf ;
               rdfs:isDefinedBy ?isDefinedBy ;
               skos:notation ?notation ;
               skos:definition ?definition ;
               nsg:atlasRelease ?atlasRelease ;
               schema:identifier ?identifier ;
               <https://sbo-nexus-delta.shapes-registry.org/ontologies/core/bmo/delineatedBy> ?delineatedBy ;
               <https://neuroshapes.org/hasLayerLocationPhenotype> ?hasLayerLocationPhenotype ;
               <https://sbo-nexus-delta.shapes-registry.org/ontologies/core/bmo/representedInAnnotation> ?representedInAnnotation ;
               <https://sbo-nexus-delta.shapes-registry.org/ontologies/core/bmo/hasLeafRegionPart> ?hasLeafRegionPart ;
               schema:isPartOf ?isPartOf ;
               <https://sbo-nexus-delta.shapes-registry.org/ontologies/core/bmo/isLayerPartOf> ?isLayerPartOf .
           }} WHERE {{
               GRAPH ?g {{
                   ?id a ?type ;
                       rdfs:label ?label ;
                   OPTIONAL {{
                   ?id rdfs:subClassOf ?subClassOf ;
                   }}
                   OPTIONAL {{
                   ?id skos:definition ?definition ;
                   }}
                   OPTIONAL {{
                   ?id skos:prefLabel ?prefLabel .
                   }}
                   OPTIONAL {{
                   ?id skos:altLabel ?altLabel .
                   }}
                   OPTIONAL {{
                   ?id rdfs:isDefinedBy ?isDefinedBy .
                   }}
                   OPTIONAL {{
                   ?id skos:notation ?notation .
                   }}
                   OPTIONAL {{
                   ?id skos:definition ?definition .
                   }}
                   OPTIONAL {{
                   ?id nsg:atlasRelease ?atlasRelease .
                   }}
                   OPTIONAL {{
                   ?id <https://neuroshapes.org/hasLayerLocationPhenotype> ?hasLayerLocationPhenotype .
                   }}
                   OPTIONAL {{
                   ?id schema:identifier ?identifier .
                   }}
                   OPTIONAL {{
                   ?id <https://sbo-nexus-delta.shapes-registry.org/ontologies/core/bmo/delineatedBy> ?delineatedBy .
                   }}
                   OPTIONAL {{
                   ?id <https://sbo-nexus-delta.shapes-registry.org/ontologies/core/bmo/representedInAnnotation> ?representedInAnnotation .
                   }}
                   OPTIONAL {{
                   ?id <https://sbo-nexus-delta.shapes-registry.org/ontologies/core/bmo/hasLeafRegionPart> ?hasLeafRegionPart .
                   }}
                   OPTIONAL {{
                   ?id schema:isPartOf ?isPartOf .
                   }}
                   OPTIONAL {{
                   ?id <https://sbo-nexus-delta.shapes-registry.org/ontologies/core/bmo/isLayerPartOf> ?isLayerPartOf .
                   }}
                   OPTIONAL {{
                   ?id <https://neuroshapes.org/units> ?units .
                   }}
                   {{
                   SELECT * WHERE {{
                       {{ ?id <https://bluebrain.github.io/nexus/vocabulary/deprecated> "false"^^xsd:boolean ; a owl:Class ;
    rdfs:subClassOf* {resource} ; rdfs:label ?label  FILTER regex(?label, {keyword}, "i") }} UNION
                       {{ ?id <https://bluebrain.github.io/nexus/vocabulary/deprecated> "false"^^xsd:boolean ; a owl:Class ;
    rdfs:subClassOf* {resource} ; skos:notation ?notation  FILTER regex(?notation, {keyword}, "i") }} UNION
                       {{ ?id <https://bluebrain.github.io/nexus/vocabulary/deprecated> "false"^^xsd:boolean ; a owl:Class ;
    rdfs:subClassOf* {resource} ; skos:prefLabel ?prefLabel  FILTER regex(?prefLabel, {keyword}, "i") }} UNION
                       {{ ?id <https://bluebrain.github.io/nexus/vocabulary/deprecated> "false"^^xsd:boolean ; a owl:Class ;
    rdfs:subClassOf* {resource} ; skos:altLabel ?altLabel  FILTER regex(?altLabel, {keyword}, "i") }}
                   }} LIMIT {search_size}
                   }}
               }}
           }}
"""  # ORDER BY ?id


async def sparql_exact_resolve(
    query: str,
    resource_type: str,
    sparql_view_url: str,
    token: str,
    httpx_client: AsyncClient,
) -> list[dict[str, str]] | None:
    """Resolve query with the knowledge graph using sparql (exact match).

    Parameters
    ----------
    query
        Query to resolve (needs to be a brain region).
    resource_type
        Type of resource to match.
    sparql_view_url
        URL to the knowledge graph.
    token
        Token to access the KG.
    httpx_client
        Async Client.

    Returns
    -------
    list[dict[str, str]] | None
        List of brain region IDs and names (only one for exact match).
    """
    # For exact match query remove punctuation and add ^ + $ for regex
    sparql_query_exact = SPARQL_QUERY.format(
        keyword=f'"^{escape_punctuation(query)}$"',
        search_size=1,
        resource=resource_type,
    ).replace("\n", "")

    # Send the sparql query
    response = await httpx_client.post(
        url=sparql_view_url,
        content=sparql_query_exact,
        headers={
            "Content-Type": "text/plain",
            "Accept": "application/sparql-results+json",
            "Authorization": f"Bearer {token}",
        },
    )
    try:
        # Get the BR or mtype ID
        object_id = response.json()["results"]["bindings"][0]["subject"]["value"]

        # Get the BR or mtype name
        object_name = next(
            (
                resp["object"]["value"]
                for resp in response.json()["results"]["bindings"]
                if "literal" in resp["object"]["type"]
            )
        )
        logger.info(
            f"Found object {object_name} id {object_id} from the exact"
            " match Sparql query."
        )

        # Return a single element (because exact match)
        return [{"label": object_name, "id": object_id}]

    # If nothing matched, notify parent function that exact match didn't work
    except (IndexError, KeyError):
        return None


async def sparql_fuzzy_resolve(
    query: str,
    resource_type: str,
    sparql_view_url: str,
    token: str,
    httpx_client: AsyncClient,
    search_size: int = 10,
) -> list[dict[str, str]] | None:
    """Resolve query with the knowledge graph using sparql (fuzzy match).

    Parameters
    ----------
    query
        Query to resolve (needs to be a brain region).
    resource_type
        Type of resource to match.
    sparql_view_url
        URL to the knowledge graph.
    token
        Token to access the KG.
    httpx_client
        Async Client.
    search_size
        Number of results to retrieve.


    Returns
    -------
    list[dict[str, str]] | None
        List of brain region IDs and names. None if none found.
    """
    # Prepare the fuzzy sparql query
    sparql_query_fuzzy = SPARQL_QUERY.format(
        keyword=f'"{query}"',
        search_size=search_size,
        resource=resource_type,
    ).replace("\n", "")

    # Send it
    response = await httpx_client.post(
        url=sparql_view_url,
        content=sparql_query_fuzzy,
        headers={
            "Content-Type": "text/plain",
            "Accept": "application/sparql-results+json",
            "Authorization": f"Bearer {token}",
        },
    )

    results = None
    if response.json()["results"]["bindings"]:
        # Define the regex pattern for br ids
        pattern = re.compile(r"http:\/\/api\.brain-map\.org\/api\/.*")

        # Dictionary to store unique objects
        objects: dict[str, str] = {}

        # Iterate over the response to extract the required information
        for entry in response.json()["results"]["bindings"]:
            # Test if the subject is of the form of a BR id or if we are looking for mtype
            subject = entry["subject"]["value"]
            if pattern.match(subject) or "braincelltype" in resource_type.lower():
                # If so, get the predicate value and see if it describes a label
                predicate = entry["predicate"]["value"]
                if predicate == "http://www.w3.org/2000/01/rdf-schema#label":
                    label = entry["object"]["value"]
                    # Append results if seen for the first time
                    if subject not in objects:
                        objects[subject] = label

        # Convert to the desired format
        results = [
            {"label": label, "id": subject} for subject, label in objects.items()
        ]
        # Output the result
        logger.info(f"Found {len(results)} objects from the fuzzy Sparql query.")
    return results


async def es_resolve(
    query: str,
    resource_type: str,
    es_view_url: str,
    token: str,
    httpx_client: AsyncClient,
    search_size: int = 1,
) -> list[dict[str, str]] | None:
    """Resolve query with the knowlegde graph using Elastic Search.

    Parameters
    ----------
    query
        Query to resolve (needs to be a brain region).
    resource_type
        Type of resource to match.
    es_view_url
        Optional url used to query the class view of the KG. Useful for backup 'match' query.
    token
        Token to access the KG.
    httpx_client
        Async Client.
    search_size
        Number of results to retrieve.

    Returns
    -------
    list[dict[str, str]] | None
        List of brain region IDs and names. None if none found.
    """
    # Match the label of the BR or mtype
    es_query = {
        "size": search_size,
        "query": {
            "bool": {
                "must": [
                    {
                        "bool": {
                            "should": [
                                {"match": {"label": query}},
                                {"match": {"prefLabel": query}},
                                {"match": {"altLabel": query}},
                            ]
                        }
                    },
                    {"term": {"@type": "Class"}},
                ]
            }
        },
    }

    # If matching a BR, add extra regex match to ensure correct form of the id
    if "brainregion" in resource_type.lower():
        es_query["query"]["bool"]["must"].append(  # type: ignore
            {"regexp": {"@id": r"http:\/\/api\.brain-map\.org\/api\/.*"}}
        )

    # Send the actual query
    response = await httpx_client.post(
        url=es_view_url,
        headers={"Authorization": f"Bearer {token}"},
        json=es_query,
    )

    # If there are results
    if "hits" in response.json()["hits"]:
        logger.info(
            f"Found {len(response.json()['hits']['hits'])} objects from the"
            " elasticsearch backup query."
        )
        # Return all of the results correctly parsed
        return [
            {"label": br["_source"]["label"], "id": br["_source"]["@id"]}
            for br in response.json()["hits"]["hits"]
        ]
    logger.info("Didn't find brain region id. Try again next time !")
    return None


async def resolve_query(
    query: str,
    sparql_view_url: str,
    es_view_url: str,
    token: str,
    httpx_client: AsyncClient,
    resource_type: Literal["nsg:BrainRegion", "bmo:BrainCellType"] = "nsg:BrainRegion",
    search_size: int = 1,
) -> list[dict[str, str]]:
    """Resolve query using the knowlegde graph, with sparql and ES.

    Parameters
    ----------
    query
        Query to resolve (needs to be a brain region or an mtype).
    sparql_view_url
        URL to the knowledge graph.
    es_view_url
        Optional url used to query the class view of the KG. Useful for backup 'match' query.
    token
        Token to access the KG.
    httpx_client
        Async Client.
    search_size
        Number of results to retrieve.
    resource_type
        Type of resource to match.

    Returns
    -------
    list[dict[str, str]] | None
        List of brain region IDs and names. None if none found.
    """
    # Create one task per resolve method. They are ordered by 'importance'
    tasks = [
        asyncio.create_task(
            sparql_exact_resolve(
                query=query,
                resource_type=resource_type,
                sparql_view_url=sparql_view_url,
                token=token,
                httpx_client=httpx_client,
            )
        ),
        asyncio.create_task(
            sparql_fuzzy_resolve(
                query=query,
                resource_type=resource_type,
                sparql_view_url=sparql_view_url,
                token=token,
                httpx_client=httpx_client,
                search_size=search_size,
            )
        ),
        asyncio.create_task(
            es_resolve(
                query=query,
                resource_type=resource_type,
                es_view_url=es_view_url,
                token=token,
                httpx_client=httpx_client,
                search_size=search_size,
            )
        ),
    ]
    # Send them all async
    resolve_results = await asyncio.gather(*tasks)

    # Return the results of the first one that is not None (in descending importance order)
    if any(resolve_results):
        return next((result for result in resolve_results if result))
    else:
        raise ValueError(f"Couldn't find a brain region ID from the query {query}")


def escape_punctuation(text: str) -> str:
    """Escape punctuation for sparql query.

    Parameters
    ----------
    text
        Text to escape punctuation from

    Returns
    -------
        Escaped text
    """
    if not isinstance(text, str):
        raise TypeError("Only accepting strings.")
    punctuation = '-()"#/@;:<>{}`+=~|.!?,'
    for p in punctuation:
        if p in text:
            text = text.replace(p, f"\\\\{p}")
    return text



"""Endpoints for agent's question answering pipeline."""

import logging
from typing import Annotated

from fastapi import APIRouter, Depends
from fastapi.responses import StreamingResponse

from neuroagent.agents import (
    AgentOutput,
    BaseAgent,
    SimpleChatAgent,
)
from neuroagent.app.dependencies import (
    get_agent,
    get_chat_agent,
    get_connection_string,
    get_user_id,
)
from neuroagent.app.routers.database.schemas import Threads
from neuroagent.app.routers.database.sql import get_object
from neuroagent.app.schemas import AgentRequest

router = APIRouter(
    prefix="/qa", tags=["Run the agent"], dependencies=[Depends(get_user_id)]
)

logger = logging.getLogger(__name__)


@router.post("/run", response_model=AgentOutput)
async def run_agent(
    request: AgentRequest,
    agent: Annotated[BaseAgent, Depends(get_agent)],
) -> AgentOutput:
    """Run agent."""
    logger.info("Running agent query.")
    logger.info(f"User's query: {request.query}")
    return await agent.arun(request.query)


@router.post("/chat/{thread_id}", response_model=AgentOutput)
async def run_chat_agent(
    request: AgentRequest,
    _: Annotated[Threads, Depends(get_object)],
    agent: Annotated[SimpleChatAgent, Depends(get_chat_agent)],
    thread_id: str,
) -> AgentOutput:
    """Run chat agent."""
    logger.info("Running agent query.")
    logger.info(f"User's query: {request.query}")
    return await agent.arun(query=request.query, thread_id=thread_id)


@router.post("/chat_streamed/{thread_id}")
async def run_streamed_chat_agent(
    request: AgentRequest,
    _: Annotated[Threads, Depends(get_object)],
    agent: Annotated[BaseAgent, Depends(get_chat_agent)],
    connection_string: Annotated[str | None, Depends(get_connection_string)],
    thread_id: str,
) -> StreamingResponse:
    """Run agent in streaming mode."""
    logger.info("Running agent query.")
    logger.info(f"User's query: {request.query}")
    return StreamingResponse(
        agent.astream(
            query=request.query,
            thread_id=thread_id,
            connection_string=connection_string,
        )
    )



"""Schemas for the chatbot."""

import datetime
import uuid
from typing import Any, Literal, Optional

from pydantic import BaseModel
from sqlalchemy import Column, DateTime, String
from sqlalchemy.orm import declarative_base

Base = declarative_base()


def uuid_to_str() -> str:
    """Turn a uuid into a string."""
    return uuid.uuid4().hex


class Threads(Base):
    """SQL table for the chatbot's user."""

    __tablename__ = "Threads"
    # langgraph's tables work with strings, so we must too
    thread_id = Column(String, primary_key=True, default=uuid_to_str)
    user_sub = Column(String, nullable=False, primary_key=True)
    vlab_id = Column(String, nullable=False)
    project_id = Column(String, nullable=False)
    title = Column(String, default="title")
    timestamp = Column(DateTime, default=datetime.datetime.now)


class ThreadsUpdate(BaseModel):
    """Class to update the conversation's title in the db."""

    title: Optional[str] = None


class ThreadsRead(BaseModel):
    """Data class to read chatbot conversations in the db."""

    thread_id: str
    user_sub: str
    vlab_id: str
    project_id: str
    title: str = "title"
    timestamp: datetime.datetime = datetime.datetime.now()


class GetThreadsOutput(BaseModel):
    """Output of the conversation listing crud."""

    message_id: str
    entity: Literal["Human", "AI"]
    message: str


class ToolCallSchema(BaseModel):
    """Tool call crud's output."""

    call_id: str
    name: str
    arguments: dict[str, Any]



"""Conversation related CRUD operations."""

import json
import logging
from typing import Annotated, Any

from fastapi import APIRouter, Depends, HTTPException
from langchain_core.messages import AIMessage, HumanMessage, ToolMessage
from langchain_core.runnables import RunnableConfig
from langgraph.checkpoint.sqlite.aio import AsyncSqliteSaver

from neuroagent.app.dependencies import get_agent_memory
from neuroagent.app.routers.database.schemas import Threads, ToolCallSchema
from neuroagent.app.routers.database.sql import get_object

logger = logging.getLogger(__name__)

router = APIRouter(prefix="/tools", tags=["Tool's CRUD"])


@router.get("/{thread_id}/{message_id}")
async def get_tool_calls(
    _: Annotated[Threads, Depends(get_object)],
    memory: Annotated[AsyncSqliteSaver | None, Depends(get_agent_memory)],
    thread_id: str,
    message_id: str,
) -> list[ToolCallSchema]:
    """Get tool calls of a specific message.
    \f

    Parameters
    ----------
    _
        Thread object returned by SQLAlchemy
    memory
        Langgraph's checkpointer's instance.
    thread_id
        ID of the thread.
    message_id
        ID of the message.

    Returns
    -------
    tool_calls
        tools called to generate a message.

    Raises
    ------
    HTTPException
        If the thread is not from the current user.
    """  # noqa: D301, D400, D205
    if memory is None:
        raise HTTPException(
            status_code=404,
            detail={
                "detail": "Couldn't connect to the SQL DB.",
            },
        )
    config = RunnableConfig({"configurable": {"thread_id": thread_id}})
    messages = await memory.aget(config)
    if messages is None:
        raise HTTPException(
            status_code=404,
            detail={
                "detail": "Message not found.",
            },
        )
    message_list = messages["channel_values"]["messages"]

    # Get the specified message index
    try:
        relevant_message = next(
            (i for i, message in enumerate(message_list) if message.id == message_id)
        )
    except StopIteration:
        raise HTTPException(
            status_code=404,
            detail={
                "detail": "Message not found.",
            },
        )

    if isinstance(message_list[relevant_message], HumanMessage):
        return []

    # Get the nearest previous message that has content
    previous_content_message = next(
        (
            i
            for i, message in reversed(list(enumerate(message_list[:relevant_message])))
            if message.content and not isinstance(message, ToolMessage)
        )
    )

    # From sub list, extract tool calls
    tool_calls: list[ToolCallSchema] = []
    for message in message_list[previous_content_message + 1 : relevant_message]:
        if isinstance(message, AIMessage):
            tool_calls.extend(
                [
                    ToolCallSchema(
                        call_id=tool["id"], name=tool["name"], arguments=tool["args"]
                    )
                    for tool in message.tool_calls
                ]
            )

    return tool_calls


@router.get("/output/{thread_id}/{tool_call_id}")
async def get_tool_returns(
    _: Annotated[Threads, Depends(get_object)],
    memory: Annotated[AsyncSqliteSaver | None, Depends(get_agent_memory)],
    thread_id: str,
    tool_call_id: str,
) -> list[dict[str, Any] | str]:
    """Given a tool id, return its output.
    \f

    Parameters
    ----------
    _
        Thread object returned by SQLAlchemy
    memory
        Langgraph's checkpointer's instance.
    thread_id
        ID of the thread.
    tool_call_id
        ID of the tool call.

    Returns
    -------
    tool_returns
        Output of the selected tool call.

    Raises
    ------
    HTTPException
        If the thread is not from the current user.
    """  # noqa: D301, D205, D400
    if memory is None:
        raise HTTPException(
            status_code=404,
            detail={
                "detail": "Couldn't connect to the SQL DB.",
            },
        )
    config = RunnableConfig({"configurable": {"thread_id": thread_id}})
    messages = await memory.aget(config)
    if messages is None:
        raise HTTPException(
            status_code=404,
            detail={
                "detail": "Message not found.",
            },
        )
    message_list = messages["channel_values"]["messages"]

    try:
        tool_output_str = next(
            (
                message.content
                for message in message_list
                if isinstance(message, ToolMessage)
                and message.tool_call_id == tool_call_id
            )
        )
    except StopIteration:
        raise HTTPException(
            status_code=404,
            detail={
                "detail": "Tool call not found.",
            },
        )
    if isinstance(tool_output_str, str):
        try:
            tool_output = json.loads(tool_output_str)
        except json.JSONDecodeError:
            raise HTTPException(
                status_code=500,
                detail={
                    "detail": "There was an error decoding the tool output.",
                },
            )

        if isinstance(tool_output, list):
            return tool_output
        else:
            return [tool_output]
    else:
        raise HTTPException(
            status_code=500,
            detail={
                "detail": (
                    "There was an error retrieving the content of the tool output."
                    " Please forward your request to the ML team."
                ),
            },
        )



"""App utilities functions."""

import logging
from typing import Any

from fastapi import HTTPException
from httpx import AsyncClient
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.ext.asyncio import AsyncEngine, create_async_engine
from starlette.status import HTTP_401_UNAUTHORIZED

from neuroagent.app.config import Settings

logger = logging.getLogger(__name__)


async def validate_project(
    httpx_client: AsyncClient,
    vlab_id: str,
    project_id: str,
    token: str,
    vlab_project_url: str,
) -> None:
    """Check user appartenance to vlab and project before running agent."""
    response = await httpx_client.get(
        f"{vlab_project_url}/{vlab_id}/projects/{project_id}",
        headers={"Authorization": f"Bearer {token}"},
    )
    if response.status_code != 200:
        raise HTTPException(
            status_code=HTTP_401_UNAUTHORIZED,
            detail="User does not belong to the project.",
        )


def setup_engine(
    settings: Settings, connection_string: str | None = None
) -> AsyncEngine | None:
    """Get the SQL engine."""
    if connection_string:
        engine_kwargs: dict[str, Any] = {"url": connection_string}
        if "sqlite" in settings.db.prefix:  # type: ignore
            # https://fastapi.tiangolo.com/tutorial/sql-databases/#create-the-sqlalchemy-engine
            engine_kwargs["connect_args"] = {"check_same_thread": False}
        engine = create_async_engine(**engine_kwargs)
    else:
        logger.warning("The SQL db_prefix needs to be set to use the SQL DB.")
        return None
    try:
        engine.connect()
        logger.info(
            "Successfully connected to the SQL database"
            f" {connection_string if not settings.db.password else connection_string.replace(settings.db.password.get_secret_value(), '*****')}."
        )
        return engine
    except SQLAlchemyError:
        logger.warning(
            "Failed connection to SQL database"
            f" {connection_string if not settings.db.password else connection_string.replace(settings.db.password.get_secret_value(), '*****')}."
        )
        return None



"""Traces tool."""

import logging
from typing import Any, Type

from langchain_core.tools import ToolException
from pydantic import BaseModel, Field

from neuroagent.tools.base_tool import (
    BaseToolOutput,
    BasicTool,
)
from neuroagent.utils import get_descendants_id

logger = logging.getLogger(__name__)


class InputGetTraces(BaseModel):
    """Inputs of the knowledge graph API."""

    brain_region_id: str = Field(
        description="ID of the brain region of interest. Can be obtained from 'resolve-entities-tool'."
    )
    etype_id: str | None = Field(
        default=None,
        description=(
            "ID of the electrical type of the cell. Can be obtained through the 'resolve-entities-tool'."
        ),
    )


class TracesOutput(BaseToolOutput):
    """Output schema for the traces."""

    trace_id: str

    brain_region_id: str
    brain_region_label: str | None

    etype: str | None

    subject_species_id: str | None
    subject_species_label: str | None
    subject_age: str | None


class GetTracesTool(BasicTool):
    """Class defining the logic to obtain traces ids."""

    name: str = "get-traces-tool"
    description: str = """Searches a neuroscience based knowledge graph to retrieve traces names, IDs and descriptions.
    Requires a 'brain_region_id' which is the ID of the brain region of interest as registered in the knowledge graph.
    Optionally accepts an e-type id.
    The output is a list of traces, containing:
    - The trace id.
    - The brain region ID.
    - The brain region name.
    - The etype of the excited cell
    - The subject species ID.
    - The subject species name.
    - The subject age.
    The trace ID is in the form of an HTTP(S) link such as 'https://bbp.epfl.ch/neurosciencegraph/data/traces...'."""
    metadata: dict[str, Any]
    args_schema: Type[BaseModel] = InputGetTraces

    def _run(self, query: str) -> list[TracesOutput]:  # type: ignore
        pass

    async def _arun(
        self,
        brain_region_id: str,
        etype_id: str | None = None,
    ) -> list[TracesOutput] | dict[str, str]:
        """From a brain region ID, extract traces.

        Parameters
        ----------
        brain_region_id
            ID of the brain region of interest (of the form http://api.brain-map.org/api/v2/data/Structure/...)
        etype
            Name of the etype of interest (in plain english)

        Returns
        -------
            list of TracesOutput to describe the trace and its metadata, or an error dict.
        """
        logger.info(f"Entering get trace tool. Inputs: {brain_region_id=}, {etype_id=}")
        try:
            # Get descendants of the brain region specified as input
            hierarchy_ids = get_descendants_id(
                brain_region_id, json_path=self.metadata["brainregion_path"]
            )
            logger.info(
                f"Found {len(list(hierarchy_ids))} children of the brain ontology."
            )

            # Create the ES query to query the KG with resolved descendants
            entire_query = self.create_query(
                brain_region_ids=hierarchy_ids, etype_id=etype_id
            )

            # Send the query to the KG
            response = await self.metadata["httpx_client"].post(
                url=self.metadata["url"],
                headers={"Authorization": f"Bearer {self.metadata['token']}"},
                json=entire_query,
            )
            return self._process_output(response.json())
        except Exception as e:
            raise ToolException(str(e), self.name)

    def create_query(
        self,
        brain_region_ids: set[str],
        etype_id: str | None = None,
    ) -> dict[str, Any]:
        """Create ES query.

        Parameters
        ----------
        brain_region_ids
            IDs of the brain region of interest (of the form http://api.brain-map.org/api/v2/data/Structure/...)
        etype
            Name of the etype of interest (in plain english)

        Returns
        -------
            dict containing the ES query to send to the KG.
        """
        # At least one of the children brain region should match.
        conditions = [
            {
                "bool": {
                    "should": [
                        {"term": {"brainRegion.@id.keyword": hierarchy_id}}
                        for hierarchy_id in brain_region_ids
                    ]
                }
            }
        ]

        # Optionally constraint the output on the etype of the cell
        if etype_id is not None:
            logger.info(f"etype selected: {etype_id}")
            conditions.append({"term": {"eType.@id.keyword": etype_id}})  # type: ignore

        # Unwrap everything into the main query
        entire_query = {
            "size": self.metadata["search_size"],
            "track_total_hits": True,
            "query": {
                "bool": {
                    "must": [
                        *conditions,
                        {
                            "term": {
                                "@type.keyword": "https://bbp.epfl.ch/ontologies/core/bmo/ExperimentalTrace"
                            }
                        },
                        {"term": {"curated": True}},
                        {"term": {"deprecated": False}},
                    ]
                }
            },
        }
        return entire_query

    @staticmethod
    def _process_output(output: Any) -> list[TracesOutput]:
        """Process output to fit the TracesOutput pydantic class defined above.

        Parameters
        ----------
        output
            Raw output of the _arun method, which comes from the KG

        Returns
        -------
            list of TracesOutput to describe the trace and its metadata.
        """
        results = [
            TracesOutput(
                trace_id=res["_source"]["@id"],
                brain_region_id=res["_source"]["brainRegion"]["@id"],
                brain_region_label=res["_source"]["brainRegion"]["label"],
                etype=(
                    res["_source"]["eType"].get("label")
                    if "eType" in res["_source"]
                    else None
                ),
                subject_species_id=(
                    res["_source"]["subjectSpecies"]["@id"]
                    if "subjectSpecies" in res["_source"]
                    else None
                ),
                subject_species_label=(
                    res["_source"]["subjectSpecies"]["label"]
                    if "subjectSpecies" in res["_source"]
                    else None
                ),
                subject_age=(
                    f"{res['_source']['subjectAge']['value']} {res['_source']['subjectAge']['unit']}"
                    if "subjectAge" in res["_source"]
                    else None
                ),
            )
            for res in output["hits"]["hits"]
        ]
        return results
